import gym
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from torch.distributions import Categorical
from collections import deque


class Policy(nn.Module):
    def __init__(self, s_size, a_size, h_size):
        super(Policy, self).__init__()
        self.fc1 = nn.Linear(s_size, h_size)
        self.fc2 = nn.Linear(h_size, a_size)
        self.drop = nn.Dropout(0.3)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = self.drop(x)
        x = self.fc2(x)
        return F.softmax(x, dim=1)

    def act(self, state):
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        state = torch.from_numpy(state).float().unsqueeze(0).to(device)
        probs = self.forward(state).cpu()
        m = Categorical(probs)
        action = m.sample()
        return action.item(), m.log_prob(action)


def reinforce(env, policy, optimizer, n_training_episodes, max_t, gamma, print_every):
    # Help us to calculate the score during the training
    scores_deque = deque(maxlen=100)
    scores = []

    for i_episode in range(1, n_training_episodes + 1):
        saved_log_probs = []
        # reset the environment
        state = env.reset()
        # reset the buffer
        rewards = []

        # repeat until the end of the episode of len = max_t
        for t in range(max_t):
            # compute action probabilities and sample action
            action, log_prob = policy.act(state)
            saved_log_probs.append(log_prob)
            # Step the environment with the action
            state, reward, done, _ = env.step(action)
            # Store rewards in our buffers
            rewards.append(reward)
            if done:
                break

        # compute sum of the rewards
        scores_deque.append(sum(rewards))
        scores.append(sum(rewards))

        # calculate the return
        returns = deque(maxlen=max_t)
        n_steps = len(rewards)

        for t in range(n_steps)[::-1]:
            disc_return_t = (returns[0] if len(returns) > 0 else 0)
            returns.appendleft(gamma * disc_return_t + rewards[t])

            # standardization of the returns is employed to make training more stable
        eps = np.finfo(np.float32).eps.item()
        # eps is the smallest representable float, which is
        # added to the standard deviation of the returns to avoid numerical instabilities
        returns = torch.tensor(returns)
        returns = (returns - returns.mean()) / (returns.std() + eps)

        # Compute the policy loss as -sum(log(prob) * return)
        policy_loss = []
        for log_prob, disc_return in zip(saved_log_probs, returns):
            policy_loss.append(-log_prob * disc_return)
        policy_loss = torch.cat(policy_loss).sum()

        # Update the policy using an Adam optimizer:
        optimizer.zero_grad()
        policy_loss.backward()
        optimizer.step()

        if i_episode % print_every == 0:
            print('Episode {}\tAverage Score: {:.2f}'.format(i_episode, np.mean(scores_deque)))

    return scores


if __name__ == "__main__":
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    env = gym.make("CartPole-v1")
    state_size = env.observation_space.shape[0]
    action_size = env.action_space.n

    cartpole_hyperparameters = {"h_size": 128,
                                "n_training_episodes": 500,
                                "n_evaluation_episodes": 10,
                                "max_t": 1000,
                                "gamma": 0.99,
                                "lr": 5e-3,
                                "state_space": state_size,
                                "action_space": action_size,
                                }

    cartpole_policy = Policy(cartpole_hyperparameters["state_space"],
                             cartpole_hyperparameters["action_space"],
                             cartpole_hyperparameters["h_size"]
                             ).to(device)

    cartpole_optimizer = optim.Adam(cartpole_policy.parameters(),
                                    lr=cartpole_hyperparameters["lr"]
                                    )

    scores = reinforce(env,
                       cartpole_policy,
                       cartpole_optimizer,
                       cartpole_hyperparameters["n_training_episodes"],
                       cartpole_hyperparameters["max_t"],
                       cartpole_hyperparameters["gamma"],
                       print_every=100
                       )

    plt.plot(scores)
    plt.xlabel("Episodes")
    plt.ylabel("Total Reward")
    plt.savefig("./plots/reinforce_reward.png")
    plt.show()